import sys

from locust import HttpUser
from locust import task
from locust.clients import HttpSession

from je_load_density.utils.get_data_strcture.get_api_data import get_api_response_data
from je_load_density.utils.exception.exception import LoadDensityAssertException

loading_test_detail_dict = dict()
record_list = list()


def create_loading_test_user(user_detail_dict: dict, **kwargs):
    """
    another_test_setting_dict are optional
    :param user_detail_dict: detail_dict should be included host http_method test_path
    """
    http_method = user_detail_dict.get("request_method", "get")
    request_url = user_detail_dict.get("request_url", "http://localhost")
    another_test_setting_dict = user_detail_dict.get("another_test_setting_dict", dict())
    assert_result_dict = user_detail_dict.get("assert_result_dict", None)
    loading_test_detail_dict.update(
        {
            "http_method": http_method,
            "request_url": request_url,
            "another_test_setting_dict": another_test_setting_dict,
            "assert_result_dict": assert_result_dict
        }
    )
    return HttpUserWrapper


def http_method_and_assert(with_httpsession: [
    HttpSession.get, HttpSession.head, HttpSession.put, HttpSession.post,
    HttpSession.patch, HttpSession.options, HttpSession.delete
], assert_result_dict: dict):
    """
    :param with_httpsession: use to get locust httpsession data
    :param assert_result_dict: use to assert data
    :return:
    """
    with with_httpsession(
            loading_test_detail_dict.get("request_url"),
            catch_response=True,
            **loading_test_detail_dict.get("another_test_setting_dict"),
    ) as response:
        response_data = get_api_response_data(response, None, None)
        for key, value in assert_result_dict.items():
            if response_data.get(key) != value:
                raise LoadDensityAssertException(
                    "value should be {right_value} but value was {wrong_value}".format(
                        right_value=value, wrong_value=response_data.get(key)
                    )
                )


class HttpUserWrapper(HttpUser):
    """
    locust httpuser use to test
    """
    host = ""
    min_wait = 5
    max_wait = 20

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__test_client = None
        self.__loading_test_detail_dict = None
        self.__http_method_dict = None

    def on_start(self):
        self.__http_method_dict = {
            "get": self.client.get,
            "put": self.client.put,
            "delete": self.client.delete,
            "post": self.client.post,
            "head": self.client.head,
            "options": self.client.options,
            "patch": self.client.patch,
        }
        self.__loading_test_detail_dict = loading_test_detail_dict
        self.__test_client = self.__http_method_dict.get(self.__loading_test_detail_dict.get("http_method"))

    @task
    def task_with_api_testka(self):
        try:
            another_test_setting_dict: dict = self.__loading_test_detail_dict.get("another_test_setting_dict")
            assert_result_dict: dict = self.__loading_test_detail_dict.get("assert_result_dict")
            if another_test_setting_dict is not None:
                if assert_result_dict is None:
                    self.__test_client(
                        self.__loading_test_detail_dict.get("request_url"),
                        **self.__loading_test_detail_dict.get("another_test_setting_dict")
                    )
                else:
                    http_method_and_assert(self.__test_client, assert_result_dict)

            else:
                if assert_result_dict is None:
                    self.__test_client(self.__loading_test_detail_dict.get("request_url"))
                else:
                    http_method_and_assert(self.__test_client, assert_result_dict)
        except Exception as error:
            print(repr(error), file=sys.stderr)



# locust error
not_found_locust_error: str = "locust not found"
# json error
cant_reformat_json_error: str = "can't reformat json is type right?"
wrong_json_data_error: str = "can't parser json"
cant_find_json_error: str = "can't find json"
cant_save_json_error: str = "can't save json"
# executor error
executor_data_error: str = "executor receive wrong data"
executor_list_error: str = "executor receive wrong data list is none or wrong type"
# HTML
html_generate_no_data_tag: str = "record is None"
# add command
add_command_exception_tag: str = "command value type should be as method or function"
# argparse
argparse_get_wrong_data: str = "argparse receive wrong data"



class LocustNotFoundException(Exception):
    pass


class LoadDensityTestException(Exception):
    pass


class LoadDensityTestJsonException(LoadDensityTestException):
    pass


class LoadDensityTestExecuteException(LoadDensityTestException):
    pass


class LoadDensityAssertException(LoadDensityTestException):
    pass


class LoadDensityHTMLException(LoadDensityTestException):
    pass


class LoadDensityAddCommandException(LoadDensityTestException):
    pass



from je_load_density.wrapper.locust_template.http_user_with_api_testka import create_loading_test_user
from je_load_density.wrapper.locust_as_library.locust_as_library import start_test


def loading_test_with_user(
        user_detail_dict: dict,
        user_count: int = 50, spawn_rate: int = 10, test_time: int = 60,
        web_ui_dict: dict = None,
        **kwargs
):
    """
    :param user_detail_dict: dict use to create user
    :param user_count: how many user we want to spawn
    :param spawn_rate: one time will spawn how many user
    :param test_time: total test run time
    :param web_ui_dict: web ui dict include host and port like {"host": "127.0.0.1", "port": 8089}
    :param kwargs: to catch unknown param
    :return: None
    """
    user = create_loading_test_user(user_detail_dict)
    start_test(
        user, user_count, spawn_rate, test_time,
        web_ui_dict, **kwargs
    )



import gevent
from locust.env import Environment
from locust.stats import stats_printer, stats_history
from locust.log import setup_logging
from locust import User
from locust import events

from je_load_density.utils.test_record.test_record_class import test_record_instance

setup_logging("INFO", None)


@events.request.add_listener
def handle_request(request_type, name, response_time, response_length, response,
                   context, exception, start_time, url, **kwargs):
    """
    handle every request event to record data
    :param request_type: default request event value
    :param name: default request event value
    :param response_time: default request event value
    :param response_length: default request event value
    :param response: default request event value
    :param context: default request event value
    :param exception: default request event value
    :param start_time: default request event value
    :param url: default request event value
    :param kwargs: catch some unknown param
    :return: None
    """
    if exception:
        test_record_instance.error_record_list.append(
                {
                    "http_method": request_type,
                    "test_url": url,
                    "name": name,
                    "status_code": response.status_code,
                    "error": exception
                 }
        )
    else:
        test_record_instance.test_record_list.append(
                {
                    "http_method": request_type,
                    "test_url": url,
                    "name": name,
                    "status_code": response.status_code,
                    "text": response.text,
                    "content": response.content,
                    "headers": response.headers,
                }
        )


def create_env(user_class: [User], another_event: events = events):
    """
    :param another_event: you can use your locust event setting but don't change locust request event
    :param user_class: locust user class
    :return: locust Environment(user_class, events) events is default event
    """
    env = Environment(user_classes=[user_class], events=another_event)
    env.create_local_runner()
    gevent.spawn(stats_printer(env.stats))
    gevent.spawn(stats_history, env.runner)
    return env


def start_test(user_class: [User], user_count: int = 50, spawn_rate: int = 10, test_time: int = 60,
               web_ui_dict: dict = None,
               **kwargs):
    """
    :param user_class: locust user class
    :param user_count: how many user we want to spawn
    :param spawn_rate: one time will spawn how many user
    :param test_time: total test run time
    :param web_ui_dict: web ui dict include host and port like {"host": "127.0.0.1", "port": 8089}
    :param kwargs: to catch unknown param
    :return: None
    """
    env = create_env(user_class)
    env.runner.start(user_count, spawn_rate=spawn_rate)
    if web_ui_dict is not None:
        env.create_web_ui(web_ui_dict.get("host", "127.0.0.1"), web_ui_dict.get("port", "8089"))
    if test_time is not None:
        gevent.spawn_later(test_time, lambda: env.runner.quit())
    env.runner.greenlet.join()
    if web_ui_dict is not None:
        env.web_ui.stop()



import sys
import types

from je_load_density.utils.exception.exception import LoadDensityTestExecuteException
from je_load_density.utils.exception.exception_tag import executor_data_error, add_command_exception_tag
from je_load_density.utils.exception.exception_tag import executor_list_error
from je_load_density.utils.html_report.html_report_generate import generate_html
from je_load_density.utils.json.json_file.json_file import read_action_json
from je_load_density.wrapper.env_with_user.wrapper_env_and_user import loading_test_with_user


class Executor(object):

    def __init__(self):
        self.event_dict = {
            "loading_test_with_user": loading_test_with_user,
            "generate_html": generate_html,
        }

    def _execute_event(self, action: list):
        """
        :param action: execute action
        :return: what event return
        """
        event = self.event_dict.get(action[0])
        if len(action) == 2:
            return event(**action[1])
        elif len(action) == 1:
            return event()
        else:
            raise LoadDensityTestExecuteException(executor_data_error + " " + str(action))

    def execute_action(self, action_list: [list, dict]) -> dict:
        """
        execute all action in action list
        :param action_list: like this structure
        [
            ["method on event_dict", {"param": params}],
            ["method on event_dict", {"param": params}]
        ]
        for loop and use execute_event function to execute
        :return: recode string, response as list
        """
        if type(action_list) is dict:
            action_list = action_list.get("load_density", None)
            if action_list is None:
                raise LoadDensityTestExecuteException(executor_list_error)
        execute_record_dict = dict()
        try:
            if len(action_list) > 0 or type(action_list) is list:
                pass
            else:
                raise LoadDensityTestExecuteException(executor_list_error)
        except Exception as error:
            print(repr(error), file=sys.stderr)
        for action in action_list:
            try:
                event_response = self._execute_event(action)
                execute_record = "execute: " + str(action)
                execute_record_dict.update({execute_record: event_response})
            except Exception as error:
                print(repr(error), file=sys.stderr)
                print(action, file=sys.stderr)
                execute_record = "execute: " + str(action)
                execute_record_dict.update({execute_record: repr(error)})
        for key, value in execute_record_dict.items():
            print(key)
            print(value)
        return execute_record_dict

    def execute_files(self, execute_files_list: list):
        """
        execute action on all file in execute_files_list
        :param execute_files_list: list include execute files path
        :return: every execute detail as list
        """
        execute_detail_list = list()
        for file in execute_files_list:
            execute_detail_list.append(self.execute_action(read_action_json(file)))
        return execute_detail_list


executor = Executor()


def add_command_to_executor(command_dict: dict):
    for command_name, command in command_dict.items():
        if isinstance(command, (types.MethodType, types.FunctionType)):
            executor.event_dict.update({command_name: command})
        else:
            raise LoadDensityTestExecuteException(add_command_exception_tag)


def execute_action(action_list: list) -> dict:
    return executor.execute_action(action_list)


def execute_files(execute_files_list: list) -> list:
    return executor.execute_files(execute_files_list)
